/*

  emonPi  Continuous Sampling, modified by me, MJS, for use in an Arduino Uno with an emonTxShield connected via UART Tx to a Raspberry Pi 2 running
  emonpi 20Feb16 image.  The Shield has no rf connectivity.
  
  This version is drastically trimmed, removing several routines and libraries that are not relevant in my emonTxShield, with no LCD , no RF, and no
  temp sensors. This saves around 7k in storage space in the Arduino.

  
  It transpires that my gas meter, a Schlumberger R5, has a magnetically operated reed switch which triggers closed for every 1 cubic foot of gas
  passed through it. Closure is about 1/4 of the full cycle of the red needle on the meter. It is debounced using a Max6816, and is connected to IRQ1
  on D3 of the TxShield, and sent as part of the data to the attached Pi. It appears to debounce spectacularly well, I am seeing zero spurious pulses!
     
  Emoncms has a nice app called mysolarpv, which displays both electrical power consumed (in blue) along with solar power generated (in yellow) within
  the same graph, and I would like a direct equivalent for electric and gas usage. However both the feeds needed in mysolar require feeds in Watts,
  and do all the rest from there. To pass my gas data into this app requires an input/feed of gas Watts, not gas WattHours as is currently the case.
  To convert my gas WattHours to gas Watts needs the WH to be divided by the duration in hours between pulses, and that is hard to retrieve in emoncms.
 
  Update: Actually, it is only hard because suitable functions to make this straightforward do not exist in the Input process functions. I find I am
  having to create new functions anyway to do even simple basics like dividing a user-settable constant by an Input value instead of the other way
  around.
  
  I am going to track the pulsegap duration here, and send it directly as a new Input to emoncms, as a complementary measurement to the existing
  pulseCount Input, as both are derived directly from the pulse ISR, and there are some distinct advantages to making the Watts calculations with the
  Input processes, as key constants can be made user-settable to accommodate other types of pulsed meter, whereas amending them in an Arduino sketch
  makes for a much clumsier experience. And having the means to make all my calculations in the processlist is more flexible, and easier to offer to
  other users who may be reluctant to mess with their Arduino firmware sketches. 
  

  Transmit values via RFM69CW radio. Not relevant, no RF.

   ------------------------------------------
  Part of the openenergymonitor.org project

  Authors: Glyn Hudson & Trystan Lea
  Builds upon JCW JeeLabs RF12 library and Arduino

  Licence: GNU GPL V3

 I have now commented out significant chunks of code that I will never need because my Shield lacks the relevant hardware, and have saved approx 7000 bytes of storage space.
 My Arduino was having an upload issue whereby the read confirmation of the write fails at around 80%. Hopefully the reduction in size will let the sketch upload successfully.
 Comments are extensive, but are never compiled, so the commented-out code is not lost, but never takes up any space in the Arduino. 
 
 Update: I think the real issue was not letting the Arduino stabilise long enough after connecting the USB lead before attempting to upload the sketch.
  
*/

/*Recommended node ID allocation
------------------------------------------------------------------------------------------------------------
-ID-  -Node Type-
0 - Special allocation in JeeLib RFM12 driver - reserved for OOK use
1-4     - Control nodes
5-10  - Energy monitoring nodes
11-14 --Un-assigned --
15-16 - Base Station & logging nodes
17-30 - Environmental sensing nodes (temperature humidity etc.)
31  - Special allocation in JeeLib RFM12 driver - Node31 can communicate with nodes on any network group
-------------------------------------------------------b--------------------------------------------------


*/

//#define emonTxV3                               // Not relevant.

                                                 // This does not follow from my reading of that link. Besides the #define says emonTxV3, not emonPiV3
                                                 // EmonLibCM which assumes 3.3V throughout. My Shield uses the 5Vcc as its ref (5.00V measured).
                                                 // May need to check EmonLibCM for 3.3V usage and amend for 5V usage. Did, no issue.
#define RF69_COMPAT 1                            // Set to 1 if using RFM69CW or 0 is using RFM12B

#include <JeeLib.h>                              // https://github.com/openenergymonitor/jeelib
#include <avr/pgmspace.h>
#include <util/parity.h>
ISR(WDT_vect) { Sleepy::watchdogEvent(); }       // Attached JeeLib sleep function to Atmega328 watchdog -enables MCU to be put into sleep mode inbetween readings to reduce power consumption.

#include <EmonLibCM.h>                           // https://github.com/openenergymonitor/EmonLibCM  Continuous Sampling version of EmonLib
// #include <OneWire.h>                          // http://www.pjrc.com/teensy/td_libs_OneWire.html
// #include <DallasTemperature.h>                // http://download.milesburton.com/Arduino/MaximTemperature/DallasTemperature_LATEST.zip

#include <Wire.h>                                // Arduino I2C library
// #include <LiquidCrystal_I2C.h>                // https://github.com/openenergymonitor/LiquidCrystal_I2C
// LiquidCrystal_I2C lcd(0x27,16,2);             // LCD I2C address to 0x27, 16x2 line display

const byte firmware_version = 666;               //firmware version x 10 e.g 10 = V1.0 / 1 = V0.1

//----------------------------emonPi Settings---------------------------------------------------------------------------------------------------------------
boolean TRUE = true;                           // These were probably declared in one of the libraries I have excluded, I declare them here since
boolean FALSE = false;                         // the uppercase forms are used frequently in this sketch. "TRUE" is a variable, while "true" is a
                                               // predefined constant in the Arduino language. It is possible this saves space.

boolean debug =                   TRUE;   

const unsigned long BAUD_RATE=    38400;         // Works just fine.

const byte Vrms_EU=               230;           // Vrms for apparent power readings (when no AC-AC voltage sample is present)
const byte Vrms_USA=              110;           // USA apparent power VRMS
const int TIME_BETWEEN_READINGS=  5;             // Time between readings (mS) but multiplied by 1000, = 5 by default.

const float Ical1=                60.606;        // (2000 turns / 33 Ohm burden), (5.00Vcc reference on emonTxShield)
const float Ical2=                60.606;
float Vcal_EU=                    267.49;        // (250V x 13) / (12.15V) = 267.49 Calibration for my UK AC-AC adapter 77DB-06-09 using actual measurements.
//const float Vcal=               260;           // Calibration for EU AC-AC adapter 77DE-06-09
const float Vcal_USA=             130.0;         // Calibration for US AC-AC adapter 77DA-10-09
boolean USA=                      FALSE; 																  

// const byte TEMPERATURE_PRECISION=  12;        // 9 (93.8ms),10 (187.5ms) ,11 (375ms) or 12 (750ms) bits equal to resolution of 0.5C, 0.25C, 0.125C and 0.0625C
const byte MaxOnewire=             6;         // maximum number of DS18B20 one wire sensors
// boolean RF_STATUS=                 0;         // Turn RF on and off. I have no RF so off it is! Don't even need the variable, now.
//-------------------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------------------


//----------------------------emonPi V3 hard-wired connections---------------------------------------------------------------------------------------------------------------
const byte LEDpin=                     9;              // emonPi LED - on when HIGH
const byte shutdown_switch_pin =       8;              // Push-to-make - Low when pressed. Not connected, I will shutdown via ssh.
const byte emonpi_GPIO_pin=            5;              // Connected to Pi GPIO 17, used to activate Pi Shutdown when HIGH. Not connected.
//const byte emonpi_OKK_Tx=              6;            // On-off keying transmission Pin - not populated by default
//const byte emonPi_RJ45_8_IO=           A6;           // RJ45 pin 8 - Analog 6 (D19) - Aux I/O
const byte emonPi_int1=                1;              // RJ45 pin 6 - INT1 - PWM - Dig 3 - default pulse count input
const byte emonPi_int1_pin=            3;              // RJ45 pin 6 - INT1 - PWM - Dig 3 - default pulse count input
//const byte emonPi_int0=                2;            // Default RFM INT (Dig2) - Can be jumpered used JP5 to RJ45 pin 7 - PWM - D2
// #define ONE_WIRE_BUS                   4               // DS18B20 Data, RJ45 pin 4
//-------------------------------------------------------------------------------------------------------------------------------------------
/*
//Setup DS128B20
OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);
byte allAddress [MaxOnewire][8];  // 8 bytes per address
byte numSensors;
*/
//-------------------------------------------------------------------------------------------------------------------------------------------

//-----RFM12B / RFM69CW SETTINGS and the Jeelib packet structure which I send to the Pi via serial connection, so I need this section--------
//
// byte RF_freq = RF12_433MHZ;         // Frequency of RF69CW module can be RF12_433MHZ, RF12_868MHZ or RF12_915MHZ, set for your module. None.
byte nodeID = 5;                       // emonpi node ID
int networkGroup = 210;

typedef struct {                       // Think this is a Jeelib function, used with PayloadTX
int power1;
int power2;
int power1_plus_2;
int Vrms;
int temp[MaxOnewire];                  // I have commented out the Onewire library.
unsigned long pulseCount;
int pulseGap;                          // new Input measurement, records the duration between most recent pulse and previous one, in seconds.
                                       // I am setting 2 hours as the historical limit. I don't care if the previous pulse was longer ago than that.
                                       // Will need to amend emonhub.conf to suit, and the Watts can be calculated from this. 
} PayloadTX;                           // create JeeLabs RF packet structure - a neat way of packaging data for RF comms
PayloadTX emonPi;
//-------------------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------------------


//Global Variables Energy Monitoring
double Vcal, vrms; 
boolean CT1, CT2, ACAC, DS18B20_STATUS;
byte CT_count, Vrms;
unsigned long last_sample=0;                        // Record millis() time of last discrete sample. Will overflow after approx 50 days.
byte flag;                                          // flag to record shutdown push button press. Not in my rig.
volatile byte pulseCount = 0;                       // Not the same pulseCount as emonPi.pulseCount, this one is used by ISR to count pulses between sample times.
unsigned long now =0;
unsigned long pulsetime=0;                          // Record time of previous interrupt pulse.
unsigned long lastTime=0;                           // Record time of previous valid interrupt pulse. Used for in-ISR calculation of gas Watts.
volatile unsigned long pulseGap=0;                  // Tracks the duration between pulses.
unsigned long pulsegaplimit = 7200000;              // 2 hours. If last pulse was more than 2 hours ago, assume it was 2 hours ago.
byte transCount = 2;                                // I am sending the emonPi.pulseGap value twice before resetting it. Feed can miss inputs!
byte numTrans = transCount;                         // Used to restore transCount

/*

 The main loop will send the last pulsegap reading every 5 seconds to emoncms, even if the boiler has been off for hours, so I need a means of
 limiting this, but risk the Feed missing an Input update.  If we have a Feed that updates every 10s, but the Input updates every 5s, and the most
 recent pulsegap value is sent only once before being cleared, then the Feed can easily miss the reading, as the Input only keeps the last value
 received, which is the one shown on the Input table. Feeds which update even less frequently will miss almost every reading! This doesn't matter
 much for electric Watts, because there are always fresh readings, but the gas watts have to wait for a pulse to arrive before the calculation
 can be made and the value is only relevant for period before the pulse. This will be why the existing pulseCount always increments and never resets,
 and all calculations derived from pulses only are a bit limited.
 
 The easiest way to fix missing updates is to set the sketch data sample time to 10 seconds, with the Feed update set at 10 seconds too. Or keep the
 sample time at the current 5 seconds but send the gas Watts data twice.
 
 And another thing!
 
 The gas Watts data is only relevant to the past of a given pulse, not the future, but that data was never written into the feed when it was relevant
 because it just didn't exist then. All the relevant datapoints were empty because the Watts values they should have contained couldn't be calculated
 until the next pulse occurred, terminating and defining the pulsegap duration that the Watts calculation depends on. But once we have that data,
 there really is no good reason at all not to update all of those relevant empty datapoints with it. Why would you not, it is real data after all!
 It just needs a suitable process.

 And rather than send the calculated gas Watts data, an interesting approach might be to send the pulsegap time itself to emoncms as a new Input, to
 complement and accompany the existing pulseCount data, both being raw pulse-related data, and allow Input processes to convert it to a flat Watts
 value, and have a special log to Feed function to retrospectively write it into every datapoint recorded by the Feed since the previous pulse.

 A good reason for doing it this way is that as the calculations need calibrated constants such as Joules per cuFt, those would benefit from being
 user-settable parameters in the Inputs GUI, instead of pre-defined in a sketch. This would allow users to set suitable constants for their cubic
 Meter gas meters, or Litres or Gallons per pulse for water or fuel metering.
 
 I am now implementing this! Update! This is now up and running.
 
 This could be a lot of datapoints to update, so it would make sense to limit pulsegap to, say, ten or fifteen minutes, essentially assuming that
 all the gas in the current cuFt was recently consumed at a higher rate, rather than being continuously consumed at a very low flat rate over the
 whole time since the last pulse, which, for very long gaps is not very realistic anyway.
 
 This is the scenario of the central heating shutting down at 22:30, with 28 seconds worth of a cuFt of gas consumed since the previous pulse, just
 a few seconds short of triggering the next pulse. The next morning, at 07:30, the CH comes back on, and 2 seconds later, the next pulse is triggered.
 From any realistic point of view, it doesn't matter how that cuFt was actually consumed, because we have zero information about the usage profile
 between pulses. We assume an average flat rate over that entire time, but we could equally well assume that it was all consumed within, say,
 the last 10 or 15 minutes, or the 10 or 15 minutes after the previous pulse, and there is no data to contradict either picture. In addition, we know
 anyway that the CH was off overnight, so assuming long periods of zero consumption is perfectly legitimate. It is also useful, because it limits how
 far back we would need to update our empty datapoints, thus helping keep our writes within reasonable limits. The only alternative is to assume the
 gas was consumed at a flat rate over the whole duration between the pulses, and that could mean a LOT of update writes. Assuming it was just a few
 minutes ago is not only just as valid, but is more realistic, and more efficient.
   
 And again, using Input processes in the Input GUI for the calculations lets users set their own max durations, so we can leave the sketch's maximum
 at something like 7200 seconds to leave plenty of room for user adjustment.  
 
 
*/                                                              																 
                                                                 
// RF Global Variables
// static byte stack[RF12_MAXDATA+4], top, sendLen, dest;        // RF variables
static char cmd;
static word value;                                               // Used to store serial input
long unsigned int start_press=0;                                 // Record time emonPi shutdown push switch is pressed

const char helpText1[] PROGMEM =                                 // Available Serial Commands if using minicom from the Pi
"\n"
"Available commands:\n"
"  <nn> i     - set node IDs (standard node ids are 1..30)\n"
"  <n> b      - set MHz band (4 = 433, 8 = 868, 9 = 915)\n"
"  <nnn> g    - set network group (RFM12 only allows 212, 0 = any)\n"
"  <n> c      - set collect mode (advanced, normally 0)\n"
"  ...,<nn> a - send data packet to node <nn>, request ack\n"
"  ...,<nn> s - send data packet to node <nn>, no ack\n"
"  ...,<n> p  - Set AC Adapter Vcal 1p = UK, 2p = USA\n"
"  v          - Show firmware version\n"
;

//-------------------------------------------------------------------------------------------------------------------------------------------
// SETUP ********************************************************************************************
//-------------------------------------------------------------------------------------------------------------------------------------------
void setup()
{
  delay(100);
/*  if (USA==TRUE)
  {
    Vcal = Vcal_USA;                                      // If a USA AC/AC adapter is being used, set calibration accordingly
    Vrms = Vrms_USA;
  }
  else
  {
*/  
    Vcal = Vcal_EU;
    Vrms = Vrms_EU;
//  }

  emonPi_startup();                                       // emonPi startup procedure, check for AC waveform and print out debug

//  if (RF_STATUS==1) RF_Setup();                         // This just hangs the Arduino, as there is no RF, so RF_STATUS is set to 0
  Serial.println(F("No RF"));
  
//  byte numSensors =  check_for_DS18B20();               // check for presence of DS18B20 and return number of sensors
 
  emonPi_LCD_Startup();                                   // don't have one. Yet.

  delay(1500);
  
  CT_Detect();
  
  serial_print_startup();

  attachInterrupt(emonPi_int1, onPulse, FALLING);    // Attach pulse counting interrupt(Dig 3 / IRQ1). My gas meter has a debounced reedswitch
                                                     // attached to this IRQ, triggering it once for every cubic foot passed. The ISR counts these,
						                          	// and also tracks the duration between pulses, which is useful data in itself.
  emonPi.pulseCount = 0;                             // Initialise Pulse Count. The ISR's count is added to this then zeroed by the main loop.
                                                     // The running total of counted pulses is sent to the Pi, after adding the IRQ's count.
  emonPi.pulseGap = 0;								 // My new Input measurement.
  EmonLibCM_number_of_channels(2);                   // number of CT channels. My TxShield has four CT ports, but I only use one.
  EmonLibCM_cycles_per_second(50);                   // frequency 50Hz or 60Hz
  EmonLibCM_datalog_period(TIME_BETWEEN_READINGS);   // period of readings in seconds 5mS is set further up the sketch

  EmonLibCM_min_startup_cycles(10);                  // number of cycles to let ADC run before starting first actual measurement
                                                     // larger value improves stability if operated in stop->sleep->start mode

  EmonLibCM_voltageCal(Vcal*(5.0/1023));             // My Arduino uses 5Vcc (5.0v measured) and my Vcal is 267.49 * (5.0/1023) = 1.307.

  EmonLibCM_currentCal(0,Ical1*(5.0/1023));          // 2000 turns / 33 Ohms burden resistor, 5.00Vcc, Ical1 = 60.606 calibration.
  EmonLibCM_currentCal(1,Ical2*(5.0/1023));          // As Ical1

  EmonLibCM_phaseCal(0,0.22);                 
  EmonLibCM_phaseCal(1,0.41);
  EmonLibCM_Init();

  Serial.println(F("Setup completed, entering MainLoop..."));
  Serial.println(F(""));
} //end setup


//-------------------------------------------------------------------------------------------------------------------------------------------
// LOOP ********************************************************************************************
//-------------------------------------------------------------------------------------------------------------------------------------------
void loop()
{
  now = millis();
/*
  if (digitalRead(shutdown_switch_pin) == 0 )
    digitalWrite(emonpi_GPIO_pin, HIGH);                // if emonPi shutdown button pressed then send signal to the Pi on GPIO 11. Will never happen.
  else
    digitalWrite(emonpi_GPIO_pin, LOW);

  if (Serial.available()){
    handleInput(Serial.read());                          // If serial input is received. It better not, it isn't connected! Only Tx is connected.
    double_LED_flash();
  }
*/

/*
  if (RF_STATUS==1){                                      // IF RF module is present and enabled then perform RF tasks
    if (RF_Rx_Handle()==1)
    {                                                     // Returns true if RF packet is received
       double_LED_flash();
    }
    send_RF();                                            // Transmit data packets if needed
  }
*/

  if (EmonLibCM_Ready())
  {

    if (EmonLibCM_ACAC) {
      emonPi.Vrms = EmonLibCM_Vrms*100;

      if (CT1) {
        emonPi.power1 = EmonLibCM_getRealPower(0);
        // emonPi.wh1 = EmonLibCM_getWattHour(0);
		
      }

/*
      if (CT2) {
        emonPi.power2 = EmonLibCM_getRealPower(1);
        // emonPi.wh2 = EmonLibCM_getWattHour(1);
      }
*/
    } else {
      emonPi.Vrms = Vrms*100;
      if (CT1) {
        emonPi.power1 = Vrms * EmonLibCM_getIrms(0);
        // emonPi.wh1 = 0;
      }
/*
      if (CT2) {
        emonPi.power2 = Vrms * EmonLibCM_getIrms(1);
      // emonPi.wh2 = 0;
      }
*/	  
    }
  }

  if ((now - last_sample) > TIME_BETWEEN_READINGS*1000)                            // ie, 5 seconds
  {
    single_LED_flash();                                                            // single flash of LED on local CT sample
    emonPi.power1_plus_2=emonPi.power1 + emonPi.power2;                            // Create power 1 plus power 2 variable

    // Serial.print(emonPi.pulseCount); Serial.print(" ");delay(5);
    // if (debug==1) {Serial.print(emonPi.power2); Serial.print(" ");delay(5);}
/*
    if (DS18B20_STATUS==1)
    {
      sensors.requestTemperatures();                                      // Send the command to get temperatures
      for(byte j=0;j<numSensors;j++) emonPi.temp[j]=get_temperature(j);
    }
*/
    if (pulseCount)                                                  // if ISR1 has counted some pulses (OK, a pulse!), update the total count, then clear the ISR's copy to zero.
    {                                                                // ISR1 will count the number of new pulses since the last sample. For my gas meter, 1 pulse at most!
      cli();                                                         // Disable interrupt just in case a pulse comes in while we are updating the count.
      emonPi.pulseCount += pulseCount;                               // For the gas meter, pulseCount will be 1, because the pulses can come only once every 30s or so
                                                                     // and the sample duration is only 10 seconds and it is reset to zero here before the next ISR
                                                                     // gets a chance to increment it.
	                                                                      
      pulseCount = 0;                                                // Reset the ISR's pulseCount.
                                                                     // If we get here, we had a pulse, so we must have a new non-zero pulseGap.  
      emonPi.pulseGap = int(pulseGap/1000);                          // pulseGap is in millis(), so convert to seconds. I created a new measurement called pulsegap.
                                                                     // emonhub.conf needs to have a corresponding config so it gets handled correctly.
      sei();                                                         // Re-enable interrupts 
    }

    //Serial.print("Ct1 watts "); Serial.print(emonPi.power1); Serial.print(" ");
    //Serial.print("Ct1 amps "); Serial.print(EmonLibCM_getIrms(0)); Serial.print("A ");
    //Serial.print("V "); Serial.print(EmonLibCM_Vrms); Serial.println(F("V"));
    //Serial.println(emonPi.temp);
  
    send_emonpi_serial();                                             // Send emonPi data to Pi serial using struct packet structure
    if (transCount)
    {
    transCount--;
    }
    else
    {
    emonPi.pulseGap = 0;                                              // It has now been sent twice, so clear it for the next pulse.  
    transCount = numTrans;                                            // Restore transCount. This should be enough to ensure that a 10 second emoncms Feed catches it. 
    }                                                                 // Longer feed intervals risk completely missing this data.
    last_sample = now;                                                // Record time of sample transmission.
	     
  } // end sample
} // end loop---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void single_LED_flash()
{
  digitalWrite(LEDpin, HIGH);  delay(50); digitalWrite(LEDpin, LOW);
}

void double_LED_flash()
{
  digitalWrite(LEDpin, HIGH);  delay(25); digitalWrite(LEDpin, LOW);
  digitalWrite(LEDpin, HIGH);  delay(25); digitalWrite(LEDpin, LOW);
}
